package es.rediris.papi;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.security.Key;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.Provider;
import java.security.PublicKey;
import java.security.interfaces.RSAPrivateCrtKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;

import javax.crypto.Cipher;

import org.bouncycastle.jce.provider.BouncyCastleProvider;

import sun.misc.BASE64Decoder;
import sun.misc.BASE64Encoder;

import org.bouncycastle.util.encoders.Base64;
/**
 * RSATool is a helper class for encrypting/decrypting messages using RSA algorithm in ECB mode.
 * Also, it uses PKCS1 padding when makes its operations with messages.<br/>
 * Keys are in PEM (PKCS#8) format, so you need to include the Bouncy Castle Library.
 *
 */
public class RSATool {
    private PublicKey rsaPublicKey;
    private Key rsaPrivateKey;
    private static int paddingSize = 11;
    /**
     * Generate an RSATool object, adding the Bouncy Castle Provider
     *
     */
    public RSATool() {
        // add the security provider
    	BouncyCastleProvider bcp = new BouncyCastleProvider();
        java.security.Security.addProvider((Provider)bcp);

        // initalice keys to null
        rsaPublicKey=null;
        rsaPrivateKey=null;
    }
    
    /**
     * Get the private key, which has should been previously read
     * @return The private key, or <tt>null</tt> if it hasn't read before
     */
	public Key getRsaPrivateKey() {
		return rsaPrivateKey;
	}

	/**
	 * Read a private key file in PEM PKCS 8 format. The file should be like:<br/>
	 * <pre>
	 * -----BEGIN RSA PRIVATE KEY-----
	 * ...
	 * -----END RSA PRIVATE KEY-----
	 * </pre>
	 * @param filename The name of the file
	 * @throws IOException
	 */
	public void readRsaPrivateKeyPEM(String filename) throws IOException {
		try {
			File keyFile = new File(filename);

			BufferedReader br = new BufferedReader(new FileReader(keyFile));
			StringBuffer keyBase64 = new StringBuffer();
			String line = br.readLine ();
			while(line != null) {
				if ( !(line.startsWith("-----BEGIN")) && !(line.startsWith("-----END"))) {
					keyBase64.append (line);
				}
				line = br.readLine ();
			}

			byte [] fileBytes = new BASE64Decoder().decodeBuffer (keyBase64.toString ());

			KeyFactory kf = KeyFactory.getInstance("RSA");
			KeySpec ks = new PKCS8EncodedKeySpec(fileBytes);
			rsaPrivateKey = (RSAPrivateCrtKey)kf.generatePrivate(ks);

		} catch (InvalidKeySpecException ex1) {
			throw new IOException("Invalid Key Spec");
		} catch (NoSuchAlgorithmException ex2) {
			throw new IOException("No Such Algorithm");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

    /**
     * Get the public key, which has should been previously read
     * @return The public key, or <tt>null</tt> if it hasn't read before
     */
	public PublicKey getRsaPublicKey() {
		return rsaPublicKey;
	}

	/**
	 * Read a public key file in PEM PKCS 8 format. The file should be like:<br/>
	 * <pre>
	 * -----BEGIN PUBLIC KEY-----
	 * ...
	 * -----END PUBLIC KEY-----
	 * </pre>
	 * @param filename The name of the file
	 * @throws IOException
	 */
	public void readRsaPublicKeyPEM(String filename) throws IOException {
		try {
			File keyFile = new File(filename);

			BufferedReader br = new BufferedReader(new FileReader(keyFile));
			StringBuffer keyBase64 = new StringBuffer();
			String line = br.readLine ();
			while(line != null) {
				if ( !(line.startsWith("-----BEGIN")) && !(line.startsWith("-----END"))) {
					keyBase64.append (line);
				}
				line = br.readLine ();
			}

			byte [] fileBytes = new BASE64Decoder().decodeBuffer (keyBase64.toString ());

			X509EncodedKeySpec ks = new X509EncodedKeySpec(fileBytes);
			KeyFactory kf = KeyFactory.getInstance("RSA");
			rsaPublicKey = kf.generatePublic(ks);
		} catch (InvalidKeySpecException ex1) {
			throw new IOException("Invalid Key Spec");
		} catch (NoSuchAlgorithmException ex2) {
			throw new IOException("No Such Algorithm");
		}
	}

	/**
	 * Encrypt a message using the private key
	 * @param data The message you want to encrypt
	 * @return The ciphered message in base 64
	 */
/*	public String encode(String data) {
		String output = "";
		try {
			Cipher rsaCipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
			rsaCipher.init(Cipher.ENCRYPT_MODE,rsaPrivateKey);
			int blockSize = 117;
			int dataSize = data.length();
			int numBlocks = dataSize / blockSize;
			int count = 0;
			while (count < numBlocks) {
				int index = count*blockSize;
				String block = data.substring(index, index+blockSize);
				byte[] rawData = block.getBytes();
				byte [] results = rsaCipher.doFinal(rawData);
				String strBlock = new String (new BASE64Encoder().encode(results));
				output += strBlock;
				count ++;
			}

			if ((dataSize % blockSize) != 0) {
				int index = count*blockSize;
				String block = data.substring(index);
				
				byte[] rawData = block.getBytes();
				byte [] results = rsaCipher.doFinal(rawData);
				String strBlock = new String(new BASE64Encoder().encode(results));
				output += strBlock;
			}
			
			return output;
		} catch (Exception e) {
			return null;
		}
		
	}*/
	
	public String encode(String data) {
		String output = "";
		try {
			Cipher rsaCipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
			rsaCipher.init(Cipher.ENCRYPT_MODE,rsaPrivateKey);
			int outputBlockSize = rsaCipher.getOutputSize(1);
			int inputBlockSize = outputBlockSize - paddingSize;
			int inputSize = data.length();
			int numBlocks = inputSize / inputBlockSize;
			byte[][] rawOutput = null;
			if ((inputSize % inputBlockSize) != 0) 
				rawOutput = new byte[numBlocks+1][outputBlockSize];
			else
				rawOutput = new byte[numBlocks][outputBlockSize];
				
			int blockCount = 0;
			while (blockCount < numBlocks) {
				int index = blockCount*inputBlockSize;
				String block = data.substring(index, index+inputBlockSize);
				byte[] rawData = block.getBytes();
				rawOutput[blockCount] = rsaCipher.doFinal(rawData);
				blockCount++;
			}

			if ((inputSize % inputBlockSize) != 0) {
				numBlocks++;
				int index = blockCount*inputBlockSize;
				String block = data.substring(index);
				byte[] rawData = block.getBytes();
				rawOutput[blockCount] = rsaCipher.doFinal(rawData);
			}
			
			byte[] fullOutput = new byte[numBlocks*outputBlockSize];

			for (int i=0; i<numBlocks;i++) {
                for (int j=0; j<outputBlockSize; j++) {
                    fullOutput[i*outputBlockSize + j] = rawOutput[i][j];
                }
			}

			//output = new String(new BASE64Encoder().encode(fullOutput));
			output = new String(Base64.encode(fullOutput));
			return output;
		} catch (Exception e) {
			return null;
		}
		
	}
	
	public String decode (String data) {
		
		String output = "";
		try {
			Cipher rsaCipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
			rsaCipher.init(Cipher.DECRYPT_MODE,rsaPublicKey);
//			byte[] decodedData = new BASE64Decoder().decodeBuffer(data);
			byte[] decodedData = Base64.decode(data);
			int inputSize = decodedData.length;
			int inputBlockSize = rsaCipher.getOutputSize(1);
			int numBlocks = inputSize / inputBlockSize;
			int resto = inputSize % inputBlockSize; 
			
			byte[][] rawOutput = null;
			if (resto != 0) 
				rawOutput = new byte[numBlocks+1][];
			else
				rawOutput = new byte[numBlocks][];
				
			int blockCount = 0;
			while (blockCount < numBlocks) {
				int index = blockCount*inputBlockSize;
				//String block = data.substring(index, index+inputBlockSize);
				//byte[] rawData = block.getBytes();
				rawOutput[blockCount] = rsaCipher.doFinal(decodedData, index, inputBlockSize);
				blockCount++;
			}

			
			if (resto != 0) {
				numBlocks++;
				int index = blockCount*inputBlockSize;
				rawOutput[blockCount] = rsaCipher.doFinal(decodedData, index, resto);
			}
			
			int totalSize = 0;
			for (int i=0; i<numBlocks;i++)
				totalSize += rawOutput[i].length;
			
			byte[] fullOutput = new byte[totalSize];

			int count =0;
			for (int i=0; i<numBlocks;i++) {
				int blockSize = rawOutput[i].length;
                for (int j=0; j<blockSize; j++) {
                    fullOutput[count] = rawOutput[i][j];
                    count++;
                }
			}
			
			output = new String(fullOutput);
		
		}
		catch (Exception e) {
			e.printStackTrace();
			return null;
		}
		
		return output;
	}
	/**
	 * Decrypt a message using the public key
	 * @param data The ciphered message in base 64
	 * @return The "plain" message
	 */
/*	public String decode(String data) {
		String output="";
		try {
			Cipher rsaCipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
			rsaCipher.init(Cipher.DECRYPT_MODE,rsaPublicKey);

			int blockSize = 174;
			int count = 0;
			int dataSize = data.length();
			int numBlocks = dataSize / blockSize;

			while (count < numBlocks) {
				int index = count*blockSize;
				String block = data.substring(index, index+blockSize);
				byte[] rawData = new BASE64Decoder().decodeBuffer(block);
				byte [] results = rsaCipher.doFinal(rawData);
				String strBlock = new String(results);
				output += strBlock;
				count ++;
			}
			
			if ((dataSize % blockSize) != 0) {
				int index = count*blockSize;
				String block = data.substring(index);
				byte[] rawData = new BASE64Decoder().decodeBuffer(block);
				byte [] results = rsaCipher.doFinal(rawData);
				String strBlock = new String(results);
				output += strBlock;
			}
			
			return output;
		} catch (Exception e) {
			return null;
		}
	}*/


}
